#: Object types for object ids
object_type = {}
object_type["null"] = 0
object_type["base"] = 1
object_type["account"] = 2
object_type["asset"] = 3
object_type["force_settlement"] = 4
object_type["committee_member"] = 5
object_type["witness"] = 6
object_type["limit_order"] = 7
object_type["call_order"] = 8
object_type["custom"] = 9
object_type["proposal"] = 10
object_type["operation_history"] = 11
object_type["withdraw_permission"] = 12
object_type["vesting_balance"] = 13
object_type["worker"] = 14
object_type["balance"] = 15
object_type["OBJECT_TYPE_COUNT"] = 16
